import re
import uuid
from datetime import datetime

def remove_hex_from_string(str):
   """Remove hex value from input string
   """
   return re.sub(r'[^\x00-\x7f]',r'', str)


def is_valid_uuid(val):
    try:
        uuid.UUID(str(val))
        return True
    except ValueError:
        return False
    
def _split_dates_gen(test_date1, test_date2, N):
    
    diff = (test_date2 - test_date1 ) / N
    for idx in range(N):
        
        # using generator function to solve problem
        # returns intermediate result
        yield (test_date1 + diff * idx)
    yield test_date2

def split_dates(test_date1, test_date2, N):
    """Split dates in N equally distanced date ranges in list
    date obj e.g.:
    test_date1 = datetime.datetime(1997, 1, 1)"""
    temp = list(_split_dates_gen(test_date1, test_date2, N))

    # using strftime to convert to userfriendly format
    res = []
    for sub in temp:
        res.append(sub.strftime("%Y-%m-%d %H:%M:%S"))
        
    return res

def get_dates_from_string(text: str) -> list[datetime]:
    """Extract dates from a given string and return datetime objects in a list"""
    matches = re.findall(r'\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}', text)
    for i, match in enumerate(matches):
        matches[i] = datetime.strptime(match, '%Y-%m-%d %H:%M:%S')
    return matches

def replace_dates_in_string(text: str, dates: list) -> str:
    new = text
    for i, match in enumerate(re.finditer(r'\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}', text)):
        new = "".join((new[:match.start()],dates[i],new[match.end():]))
    return new


#-*- coding: utf-8 -*-
'''This module contains the python communication functions''' 


from ctypes import*
import ctypes
import numpy as np
import os


class ConnectGIns():
    def __init__(self):
			# Load DLL into memory, take care to have the corresponding dll to the computer operating system
        try:
            #chemin=os.path.abspath(".\\Giutility\\giutility_x64.dll")
            chemin = "libGInsUtility.so" #should only work on linux machines located in cloud stack network
            #gins_lib = "/home/gins/data/libGInsUtility.so"
            self.GINSDll = ctypes.cdll.LoadLibrary(chemin)
            print("64 bit DLL imported")
            print(chemin)
        except OSError:
            print("can not imported libGInsUtility.so")

		#if linux uncomment before and use
		#self.GINSDll = ctypes.cdll.LoadLibrary("PyQStationConnect_2_0\\libGInsData.so")
		#print("Linux System")
        #function prototypes
        self.GINSDll._CD_eGateHighSpeedPort_Init.argtypes = [c_char_p,c_int,c_int,c_int,POINTER(c_int),POINTER(c_int)]
        self.GINSDll._CD_eGateHighSpeedPort_SetBackTime.argtypes = [c_int, c_double]
        self.GINSDll._CD_eGateHighSpeedPort_InitBuffer.argtypes=[c_int,c_int,c_int]
        self.GINSDll._CD_eGateHighSpeedPort_DecodeFile_Select.argtypes = [POINTER(c_int), POINTER(c_int), c_char_p]
        self.GINSDll._CD_eGateHighSpeedPort_GetDeviceInfo.argtypes = [c_int, c_int, c_int, POINTER(c_double), c_char_p]
        self.GINSDll._CD_eGateHighSpeedPort_ReadBufferToDoubleArray.argtypes = [c_int, POINTER(c_double), c_int, c_int,POINTER(c_int), POINTER(c_int), POINTER(c_int)]
        self.GINSDll._CD_eGateHighSpeedPort_WriteOnline_Single_Immediate.argtypes = [c_int, c_int, c_double]
        self.GINSDll._CD_eGateHighSpeedPort_WriteOnline_Single.argtypes = [c_int, c_int, c_double]
        self.GINSDll._CD_eGateHighSpeedPort_WriteOnline_ReleaseOutputData.argtypes = [c_int]
        self.GINSDll._CD_eGateHighSpeedPort_Close.argtypes = [c_int, c_int]
        """"  parameters for Init connection """
        self.controllerIP=0#controllerIP.encode('UTF-8')
        self.timeout=5
        self.HSP_BUFFER=2#  1 for online; 2 for buffered values
        self.HSP_ONLINE=1#  1 for online; 2 for buffered values
        self.sampleRate=100
        # parameters for file decoding
        self.FilePath=0# = FilePath.encode('UTF-8')
        self.FileDecodeComplete=False
        #general used parameters
        self.HCLIENT=c_int(-1)
        self.HCONNECTION=c_int(-1)
        #parameters for Init buffer
        self.bufferindex=0
        self.autoRun=0
        #parameters to empty the circular buffer
        self.backtime=0
        #parameters to read information from devices
        self.location=10
        self.Adress=11
        self.SampleRate=16
        self.SerialNumber=15
        self.ChannelCount=18
        self.Channel_InfoName=0
        self.Channel_Unit=1
        self.info=c_double(0)
        self.ret=0
        self.char=ctypes.create_string_buffer(30)
        #parameters to read buffer
        
        
    def init_connection(self,controllerIP):
        self.controllerIP=controllerIP.encode('UTF-8')
        """Initialisation of the connection to a controller"""
        ret=self.GINSDll._CD_eGateHighSpeedPort_Init(self.controllerIP,self.timeout,self.HSP_BUFFER,self.sampleRate,byref(self.HCLIENT),byref(self.HCONNECTION))
        if(ret!=0):
            print("Init Connection Failed - ret:",ret)
            self.ret=ret
            return False

        #Init buffer (this is mainly to select a certain buffer by index)
        ret=self.GINSDll._CD_eGateHighSpeedPort_InitBuffer(self.HCONNECTION.value,self.bufferindex,self.autoRun)
        if(ret!=0):
            print("Init Buffer Failed - ret:",ret)
            return False
        #empty the circular buffer to get only actual data
        ret=self.GINSDll._CD_eGateHighSpeedPort_SetBackTime(self.HCONNECTION.value,self.backtime)
        if(ret!=0):
            print("SetBackTime Failed - ret:",ret)
            return False
        print("Connection initialized. IP: ", controllerIP)
        return True
    
    
    def init_online_connection(self,controllerIP):
        self.controllerIP=controllerIP.encode('UTF-8')
        """Initialisation of the connection to a controller"""
        ret=self.GINSDll._CD_eGateHighSpeedPort_Init(self.controllerIP,self.timeout,self.HSP_ONLINE,self.sampleRate,byref(self.HCLIENT),byref(self.HCONNECTION))
        if(ret!=0):
            print("Init Connection Failed - ret:",ret)
            self.ret=ret
            return False

        #Init buffer (this is mainly to select a certain buffer by index)
        ret=self.GINSDll._CD_eGateHighSpeedPort_InitBuffer(self.HCONNECTION.value,self.bufferindex,self.autoRun)
        if(ret!=0):
            print("Init Buffer Failed - ret:",ret)
            return False
        #empty the circular buffer to get only actual data
        ret=self.GINSDll._CD_eGateHighSpeedPort_SetBackTime(self.HCONNECTION.value,self.backtime)
        if(ret!=0):
            print("SetBackTime Failed - ret:",ret)
            return False
        print("Connection initialized. IP: ", controllerIP)
        return True
    
    
    def init_file(self,FilePath):
        """Initialisation of the dat-file"""
        self.FilePath= FilePath.encode('UTF-8')
        ret=self.GINSDll._CD_eGateHighSpeedPort_DecodeFile_Select(byref(self.HCLIENT),byref(self.HCONNECTION),self.FilePath)
        if ret==0:
            print("File Load OK!", self.FilePath.decode('UTF-8'))
            return True
        else:
            print("Error Loading File ", self.FilePath.decode('UTF-8'))
            return False

    def read_serial_number(self):
        """Read the serial number of the connected device"""
        ret=self.GINSDll._CD_eGateHighSpeedPort_GetDeviceInfo(self.HCONNECTION.value,self.SerialNumber,0,self.info,None)
        if(ret==0):
            print("controller serial number", self.info.value)
            return self.info.value
        else:
            print("error reading serial number!")
            return 0
                
    def read_sample_rate(self):
        """Read a buffer sampling rate"""
        ret=self.GINSDll._CD_eGateHighSpeedPort_GetDeviceInfo(self.HCONNECTION.value,self.SampleRate,0,self.info,None)
        if(ret==0):
            print("controller sample rate", self.info.value)
            return self.info.value
        else:
            print("Error reading sample rate!")
            return 0
        
    def read_channel_count(self):
        """Count the number of channels connected to a controller"""
        ret=self.GINSDll._CD_eGateHighSpeedPort_GetDeviceInfo(self.HCONNECTION.value,self.ChannelCount,0,self.info,None)
        if(ret==0):
            #print("controller channel count", self.info.value)
            return self.info.value
        else:
            print("Error reading channel count!")
            return ""
    
    def read_controller_name(self):
        """Read a controller name"""
        #p=ctypes.create_string_buffer(30)#this function works for python 3.6, for lower version the name of functions to create mutable character buffer is different
        ret=self.GINSDll._CD_eGateHighSpeedPort_GetDeviceInfo(self.HCONNECTION.value,self.location,0,None,self.char)
        if(ret==0):
            print("controller name", self.char.value.decode('UTF-8'))
            return self.char.value.decode('UTF-8')
        else:
            print("Error reading controller name!")
            return ""
        
    def read_controller_address(self):
        """Read the adress of a connected controller"""
        ret=self.GINSDll._CD_eGateHighSpeedPort_GetDeviceInfo(self.HCONNECTION.value,self.Adress,0,None,self.char)
        if(ret==0):
            print("controller adress", self.char.value.decode('UTF-8'))
            return self.char.value.decode('UTF-8')
        else:
            print("Error reading controller address")
            return ""
        
    def read_channel_names(self):
        """Read the channel name and corresponding index"""
        i=0
        self.GINSDll._CD_eGateHighSpeedPort_GetDeviceInfo(self.HCONNECTION.value,self.ChannelCount,0,self.info,None)
        ChannelNb=self.info.value
        while i<ChannelNb:
            self.GINSDll._CD_eGateHighSpeedPort_GetDeviceInfo(self.HCONNECTION.value,self.Channel_InfoName,i,None,self.char)
            print("Controller index:",i," channel name:", self.char.value.decode('UTF-8'))
            i+=1
    def read_channels_unit(self):
        """Read the channel name and corresponding index"""
        i=0
        self.GINSDll._CD_eGateHighSpeedPort_GetDeviceInfo(self.HCONNECTION.value,self.ChannelCount,0,self.info,None)
        ChannelNb=self.info.value
        while i<ChannelNb:
            self.GINSDll._CD_eGateHighSpeedPort_GetDeviceInfo(self.HCONNECTION.value,self.Channel_Unit,i,None,self.char)
            try:
                print("Controller index:",i," channel unit:", self.char.value.decode('UTF-8'))
            except UnicodeDecodeError:
                print("Controller index:",i," channel unit:", "NAN")
            i+=1        
         
    def read_index_unit(self,IndexNb):
        """Read the channel name corresponding to an index"""
        ret=self.GINSDll._CD_eGateHighSpeedPort_GetDeviceInfo(self.HCONNECTION.value,self.Channel_Unit,IndexNb,None,self.char)
        if(ret==0):
            try:
                return(self.char.value.decode('UTF-8'))
            except UnicodeDecodeError:
                # we decode °C
                if self.char.value == b'\xb0C':
                    return (u'\u00b0C')
                # we decode
                if self.char.value == b'\xb5m/m':
                    return (u'\u00b5m/m')
                #we decode
                
                else:
                    return("Unit")
        else:
            print("Error reading channel unit, index",IndexNb)
            return ""
           
    def read_index_name(self,IndexNb):
        """Read the channel name corresponding to an index"""
        ret=self.GINSDll._CD_eGateHighSpeedPort_GetDeviceInfo(self.HCONNECTION.value,self.Channel_InfoName,IndexNb,None,self.char)
        if(ret==0):
            return(self.char.value.decode('UTF-8'))
        else:
            print("Error reading channel name, index",IndexNb)
            return ""
        
    def write_online_value(self,IndexNb,WriteValue):
        """Write a single double value to a specific channel"""
        ret=self.GINSDll._CD_eGateHighSpeedPort_WriteOnline_Single_Immediate(self.HCONNECTION.value,IndexNb,WriteValue)
        if(ret==0):
            print("Value:",WriteValue,"Was added to index:",IndexNb)
            return True
        else:
            print("Could not writte value:",WriteValue,"to index:",IndexNb)
            return False
 
    
    def write_single(self,IndexNb,WriteValue):
        """Writte a single double value to a specific channel"""
        ret=self.GINSDll._CD_eGateHighSpeedPort_WriteOnline_Single(self.HCONNECTION.value,IndexNb,WriteValue)
        if(ret==0):
            print("Value",WriteValue,"was written")
            return ""
        else:
            print("Error writting channel value, index")
            return ""
    def relase_output(self):
       """Writte a single double value to a specific channel"""
       ret=self.GINSDll._CD_eGateHighSpeedPort_WriteOnline_ReleaseOutputData(self.HCONNECTION.value)
       if(ret==0):
           print("Value was released")
           return ""
       else:
           print("Error releasing data")
           return ""   
    
    #def read_buffer_frame(self):
    #    self.GINSDll._CD_eGateHighSpeedPort_GetBufferFrames.argtypes=[c_int,c_int]
    #    self.GINSDll._CD_eGateHighSpeedPort_GetBufferFrames(self.HCONNECTION.value,self.HCLIENT.value)
    #    print("buffer frame number:",)
        
    
    def yield_buffer(self,NbFrames=int(100000),fillArray=0):

        self.GINSDll._CD_eGateHighSpeedPort_GetDeviceInfo(self.HCONNECTION.value,self.ChannelCount,0,self.info,None)
        ChannelNb=int(self.info.value)
        valuesPtr=(c_double*(NbFrames*ChannelNb))()
        ReceivedFrames=c_int(0)#pointer
        ReceivedChannels=c_int(0)#pointer
        ReceivedComplete=c_int(0)#pointer
        ret=0
        while(ret==0):
            ret=self.GINSDll._CD_eGateHighSpeedPort_ReadBufferToDoubleArray(self.HCONNECTION.value,valuesPtr,(NbFrames*ChannelNb),fillArray,ReceivedFrames,ReceivedChannels,ReceivedComplete)
            chcnt=ReceivedChannels.value
            BUF=valuesPtr[0:chcnt*ReceivedFrames.value]
            buffer=np.reshape(BUF,(ReceivedFrames.value,chcnt))
            yield buffer


    def init_buffer_conn(self,sbufferID):
        self.sbufferID=sbufferID.encode('UTF-8')
        ret=self.GINSDll._CD_eGateHighSpeedPort_Init_PostProcessBuffer(self.sbufferID,byref(self.HCLIENT),byref(self.HCONNECTION))
        if(ret!=0):
            print("Init Connection Failed - ret:",ret)            
            
    def close_connection(self):
        self.GINSDll._CD_eGateHighSpeedPort_Close(self.HCONNECTION.value,self.HCLIENT.value)




#This function is needeed for file saved with sampling time>1kHZ
def read_again_buffer(buffer):
    read_next_buffer=next(buffer)
    new_buffer_values= read_next_buffer[:,:]
    return (new_buffer_values)


def read_gins_dat(connection):    
    #Call function to store buffer into the variable buffer
    buffer=connection.yield_buffer()
    #We read the dat file and save it into the variable disp
    readbuffer=next(buffer) 
    Logged_file = readbuffer[:,:]        
    #This function is needed for reading the full dat file with several buffer frames    
    while True:
        #print('we enter while')
        new_disp=read_again_buffer(buffer)
        try:
            new_time=new_disp[0,0]
        except:
            break
        last_time=Logged_file[-1,0]
        if new_time>last_time:
            Logged_file=np.vstack((Logged_file,new_disp))
        else:
            break
    print ('dat file was read')
    return Logged_file
	
def create_list_channel(connection):
    '''Go through the buffer and return some list of channels names or unit'''
    Nb=0
    list_channels=[]
    list_unit=[]
    while Nb<int(connection.read_channel_count()):
        try:
            #list_channels.append(connection.read_index_name(Nb).encode('UTF-8'))
            list_channels.append(str(connection.read_index_name(Nb)))
        except AttributeError:
            list_channels.append(str(connection.read_index_name(Nb)))
        list_unit.append(connection.read_index_unit(Nb).lstrip())      
        Nb+=1
    return (list_channels,list_unit)



"""
@author: aljo, 
Module to send simplified http request to the Cloud
"""

import csv
from io import BytesIO
import requests
import datetime as dt
import numpy as np
import pandas as pd
import json 
import re
import uuid
import time 
import logging

from dataclasses import dataclass
from typing import Type, List, Dict, Tuple, Optional, Union, Any
from requests.auth import HTTPBasicAuth 
from enum import Enum
from dateutil import tz

from gimodules.cloudconnect import utils

# Set output level to INFO because default is WARNNG
logging.getLogger().setLevel(logging.INFO)

@dataclass
class CsvConfig:
    '''Object for tracking parameters for csv import'''
    # csv config - initialised with default values
    ColumnSeparator: str = ";"
    DecimalSeparator: str= ","
    NameRowIndex: int = 0
    UnitRowIndex: int = 0
    ValuesStartRowIndex: int = 1
    ValuesStartColumnIndex: int = 1
    ## Column 1: Date and Time -> specified in Gantner http docs
    ## Comment one out: if python formatter differs from C++ formatter
    ## e.g. "%d.%m.%Y %H:%M:%S.%F" on backend -> "%d.%m.%Y %H:%M:%S.%f" for python
    DateTimeFmtColumn1: str = "%d.%m.%Y %H:%M:%S.%F"
    DateTimeFmtColumn2: str = ""
    DateTimeFmtColumn3: str = ""
    
    def get_config(self):
        """returns config as dict"""
        return {
            "ColumnSeparator": self.ColumnSeparator,
            "DecimalSeparator": self.DecimalSeparator,
            "NameRowIndex": self.NameRowIndex,
            "UnitRowIndex": self.UnitRowIndex,
            "ValuesStartRowIndex": self.ValuesStartRowIndex,
            "ValuesStartColumnIndex": self.ValuesStartColumnIndex,
            "DateTimeFmtColumn1": self.DateTimeFmtColumn1,
            "DateTimeFmtColumn2": self.DateTimeFmtColumn2,
            "DateTimeFmtColumn3": self.DateTimeFmtColumn3
        }

@dataclass()
class GIStreamVariable: 
    '''Object for tracking available variables'''
    id: str
    name: str
    index: str
    unit: str
    data_type: str
    sid: str
        
class CloudRequest():
    
    def __init__ (self):
        self.url = ""
        self.user = ""
        self.pw = ""
        self.login_token = None
        self.refresh_token = None
        self.streams = None
        self.stream_variabels = None
        self.query = ""
        self.request_measurement_res = None

        # enums 
        self.resolutions = Resolution 
        self.units = None

        # importer data
        self.import_session_res_udbf = None
        self.import_session_res_csv = None
        self.import_session_csv_current = None
        self.session_ID = None
        self.csv_config = CsvConfig()


    def login(self, url:str, user:str, password:str):

        self.url = url 
        self.user = user 
        self.pw = password

        # prepare request
        login_form = { 'username': self.user, 'password': self.pw, 'grant_type': 'password' }
        auth = HTTPBasicAuth('gibench', '')
        headers = {'Content-Type': 'application/x-www-form-urlencoded'}
        login_url = self.url + '/token'

        # send request
        try:
            res = requests.post(login_url, data = login_form, headers = headers, auth = auth)
            if res.status_code == 200:
                self.login_token = res.json()
                self.refresh_token = self.login_token['refresh_token']
                logging.info("Login successful")
                self.get_all_stream_metadata()
                self.print_streams()
                self.get_all_var_metadata()
            else: 
                logging.error(f"Login failed! \nResponse Code:{res.status_code} \nReason: {res.reason}")
        except Exception as e:
            logging.warning(e)
    
    def refresh_access_token(self): 
        """
        Refresh the access token with a refresh token. 
        The refresh token is valid for 14 days. 
        """
        # prepare request
        refresh_form = json.dumps({ 'ClientID': 'gibench', 'RefreshToken': self.refresh_token })
        headers = {'Content-Type': 'application/json'}
        refresh_url = self.url + '/rpc/AdminAPI.RefreshToken'
        
        # send request
        try: 
            res = requests.post(refresh_url, data = refresh_form, headers= headers)
            if res.status_code == 200:
                res = res.json()
                res['access_token'] = res.pop('AccessToken')
                self.login_token = res
                logging.info("Access Token refreshed")
            else: 
                logging.error(f"Could not fetch access token with refresh token\nResponse Code:{res.status_code} \nReason: {res.reason}")
        except Exception as e:
            logging.warning(e)
            
        
    def get_all_stream_metadata(self): 
        """Loads the available meta information from all available streams. 
        The data is stored in data classes (GIStream) and is accessible via the .streams attribute. 
        """
        # check that auth_token exist
        if self.login_token is not None: 
            # prepare request
            url_list = self.url + '/kafka/structure/sources'
            headers = {'Authorization': 'Bearer ' + self.login_token['access_token']}
            # send request
            try: 
                res = requests.get(url_list, headers=headers)
                if res.status_code == 200: 
                    res = res.json()
                    self.streams = {} # reset memory 
                    for i in range(0, len(res['Data'])):
                        name = res['Data'][i]["Name"]
                        id = res['Data'][i]["Id"]
                        sample_rate_hz = res['Data'][i]["SampleRateHz"]
                        first_ts = res['Data'][i]["AbsoluteStart"]
                        last_ts = res['Data'][i]["LastTimeStamp"]
                        index = res['Data'][i]["Index"]
                        # store data in dict with dataclass
                        self.streams[name] = GIStream(name, id, sample_rate_hz, first_ts, last_ts, index)
                elif res.status_code == 401 or res.status_code == 403: 
                    self.refresh_access_token()
                    self.get_all_stream_metadata()
                else: 
                    logging.error(f"failed! \nResponse Code: {res.status_code} \nReason: {res.reason}")
            except Exception as e:
                logging.warning(e)
        else: 
            logging.error("You have no valid access token! \nPlease login first.")

    def print_streams(self):
        if self.streams is not None:
            for s in self.streams: 
                logging.info(f"{'Streamname:':<20}{self.streams[s].name} \
                        \n{'Streamid:':<20}{self.streams[s].id} \
                        \n{'first ts:':<20}{self.streams[s].first_ts} {dt.datetime.utcfromtimestamp(float(self.streams[s].first_ts)/1000).strftime('%Y-%m-%d %H:%M:%S')} \
                        \n{'last ts:':<20}{self.streams[s].last_ts} {dt.datetime.utcfromtimestamp(float(self.streams[s].last_ts)/1000).strftime('%Y-%m-%d %H:%M:%S')} \
                        \n{'Samplerate:':<20}{self.streams[s].sample_rate_hz}\n")
        else: 
            logging.info("You have no loaded Streams")

    def get_all_var_metadata(self):
        """Loads the available meta information from all available variable. The data is stored in data classes and is accessible via the .stream_variabels attribute. 
        """

        if self.streams is not None: 
            url_list=self.url+'/__api__/gql'
            headers = {'Authorization': 'Bearer ' + self.login_token["access_token"]}
            self.stream_variabels = {} # reset memory
            unit_names = []
            for s in self.streams:
                query="{\n  variableMapping(sid: \""+ self.streams[s].id +"\") {\n    sid\n    columns {\n      name\n      variables {\n        id\n        dataType\n        name\n        unit\n      }\n    }\n  }\n}"
                try:
                    res = requests.post(url_list, json = {'query': query}, headers = headers)
                    if res.status_code == 200: 
                        res = res.json()
                        for i in range(0, len(res['data']['variableMapping']['columns'])):
                            sid = res['data']['variableMapping']['sid']     
                            index = res['data']['variableMapping']['columns'][i]['name'] # Var index in GI System - needed for gql queries
                            name = res['data']['variableMapping']['columns'][i]['variables'][0]['name'] # Var name defined by user
                            data_type = res['data']['variableMapping']['columns'][i]['variables'][0]['dataType']
                            id = res['data']['variableMapping']['columns'][i]['variables'][0]['id']
                            unit = res['data']['variableMapping']['columns'][i]['variables'][0]['unit']

                            # create Enum for filtering units
                            if unit not in unit_names and unit != "": 
                                unit_names.append(unit)
                            # Create unique variable name /with stream
                            # TODO - if multiple variables with same name exists (in one stream) the name will be overwritten
                            self.stream_variabels[f"{s}__{name}"] = GIStreamVariable(id, name, index, unit, data_type, sid)
                    elif res.status_code == 401 or res.status_code == 403: 
                        self.refresh_access_token()
                        self.get_all_var_metadata()
                    else: 
                        logging.error(f"Fetching variable info failed! \nResponse Code:{res.status_code} \nReason: {res.reason}")
                except Exception as e:
                    logging.warning(e)
            # create Enum for available units
            self.units = Enum('Units', list(unit_names))
        else: 
            logging.info("You have no loaded Streams. \nPlease load first Streams.")

    def variable_info(self):
        '''use this endpoint to read information for all available online variables'''
        url_list = self.url + '/online/structure/variables'
        headers = {'Content-Type':'application/json','Authorization': 'Bearer ' + self.login_token["access_token"]}
        try:
            res = requests.get(url_list, headers = headers)
            if res.status_code == 200: 
                res = res.json()
                ## TODO continue here
            else: 
                logging.error(f"Fetching variable info failed! \nResponse Code:{res.status_code} \nReason: {res.reason}")
        except Exception as e:
            logging.warning(e)
            
    def find_var(self, var_name: str): 
        """Searches the existing variables. 

        Args:
            var_name (str): Var_name or Substring

        Returns:
            Dict ([str,GIStreamVariable]): List of variables found
        """

        if type(var_name) == list: 
            match = []  
            for k in self.stream_variabels.keys():
                for vr in var_name: #TODO Aviod Big O notation
                    if vr in k: 
                        match.append(k)
        else: 
            match = [x for x in self.stream_variabels.keys() if var_name in x]

        if len(match) == 0 and match is not None: 
            logging.info("No var found.")
        else:
            result = {}  
            for m in match:
                result[m] = (self.stream_variabels[m])
            return result
        return None

    def filter_var_attr(self, attr: str, value: str): 
        """
        Search for stream_variables with a certain attribute

        Args:
            attr (str): e.g. unit
            value (str): e.g. C°

        Returns:
            dataclass (obj): List of variables found
        """

        if self.stream_variabels is not None: 
            match = [x for x in self.stream_variabels.values() if getattr(x, attr, "") == value]
            return match
        return None
    
    def _build_sensorid_querystring(self, indices:list, aggregations=None):
        #TODO - enable multiple aggregations
        agg = """
                avg
        """
        s = ""
        for i in indices:
            s = s+f"""
                {i} {{
                    {agg}
                }}
            """
        return s

    def get_var_data(self, sid:str, index_list:list, start_date:str, end_date:str, resolution:str = 'nanos'):
        """Returns a np.matrix of data and pandas df with timestamps and values directly from a data stream

        Args:
            sid (str): stream_id
            index_list (list): channel_index e.g. ["a10", "a11"]
            start_date (str): format: YYYY-MM-DD HH:MM:SS
            end_date (str): format: YYYY-MM-DD HH:MM:SS
            resolution (str, optional): 'MONTH','WEEK','DAY','HOUR','QUARTER_HOUR','MINUTE','SECOND','HZ10','HZ100','KHZ','KHZ10','nanos'
                The available resolutions are accessible via the attribute ".resolutions". 
                Do not forget at the end the .values  Defaults to 'nanos'.

        Returns:
            pandas.df: dataframe
        """

        tss = str(self.convert_datetime_to_unix(start_date))
        tse = str(self.convert_datetime_to_unix(end_date))

        selected_index_string="" 
        if resolution=='nanos' and len(index_list) > 0: # nanos == raw data
            for j in range(0,len(index_list)): # concat var_index for query
                selected_index_string = selected_index_string + "\"" + index_list[j] + "\"" + " ,"
            # build query 
            
            self.query ="{\n  Raw(columns: [\"ts\", \"nanos\"," + selected_index_string + "], \
                sid: \"" + sid + "\",\
                from: " + tss + ",\
                to: " + tse + ") {\n    data\n  }\n}\n"             
        
        elif len(index_list) > 0:
            selected_index_string = self._build_sensorid_querystring(index_list)

            self.query = f"""
            {{
                analytics(
                    from: {tss},
                    to: {tse},
                    resolution: {resolution},
                    sid: "{sid}"
                ) {{
                    ts
                    {selected_index_string}
                }}
            }}
        """
        else:
            logging.info("no variable selected")

        url_list = self.url+'/__api__/gql'
        headers = {'Authorization': 'Bearer ' + self.login_token["access_token"]}
        try:
            res = requests.post(url_list, json={'query':self.query}, headers = headers)
            if res.status_code == 200 and not "errors" in res.text: 
                requested_data = res.json()

                ### Filter Data out of request
                if resolution=='nanos':
                    data_matrix=requested_data['data']['Raw']['data']
                    self.data=np.array(data_matrix)
                else:
                    self.data=np.zeros((len(requested_data['data']['analytics']['ts']),len(index_list)+1))
                    self.data[:,0]=requested_data['data']['analytics']['ts']
                    for k in range(0,len(index_list)):
                        self.data[:,k+1]=requested_data['data']['analytics'][index_list[k]]['avg']
                
                ### create numpy matrix 
                self.data=self.data.astype(float)
                nonecheck=np.isnan(self.data[:,:])
                none_index=np.where(nonecheck == True)

                #none number can be at the begining or at the end.
                if len(none_index[0])>0:
                    first_none=none_index[0][0]
                    last_none=none_index[0][-1]
                    if last_none==len(self.data)-1:
                        self.data=self.data[0:first_none-1,:]
                    else:    
                        self.data=self.data[last_none+1:-1,:]

                ### create pandas df
                self.df = pd.DataFrame(self.data, columns=self.__get_column_names(sid, index_list))
                self.df['Time'] = pd.to_datetime(self.df['Time'], unit='ms')
                return self.df 
            elif res.status_code == 401 or res.status_code == 403:
                logging.info("Token expired. Renewing...")
                self.refresh_access_token()
                return self.get_var_data(sid, index_list, start_date, end_date, resolution)
            # Show error Message from Server
            else: 
                error = json.loads(res.text)
                logging.error(f"Fetching Data failed! \nResponse Code: {res.status_code} \nReason: {res.reason}\nMsg: {error['errors'][0]['message']}")
        except Exception as e:
            logging.warning(e)
    
    def _get_stream_name_for_sid_vid(self, sid:str, vid:str):
        if self.stream_variabels != None:
            stream = [k for k, v in self.stream_variabels.items() if (v.sid == sid and v.id == vid)]
            if len(stream) == 1:
                return stream[0].split('__')[0]
        else: 
            logging.info("no stream_variables available")
            return None
        
    def _get_stream_name_for_sid(self, sid:str):
        if self.stream_variabels != None:
            stream = [stream_name for stream_name,gi_stream in self.streams.items() if gi_stream.id == sid]
            if len(stream) == 1:
                return stream[0]
        else: 
            logging.info("no stream_variables available")
            return None
    
    def get_all_vars_of_stream(self, sid:str):
        if self.stream_variabels != None:
            variables = [v for v in self.stream_variabels.values() if v.sid == sid]
            return variables
        else: 
            logging.info("no stream_variables available")
            return None
        
    def get_data_as_csv(self, variables: list[GIStreamVariable], resolution: str, start: str, end: str, filepath: str='', streaming: bool = True, return_df: bool= True, write_file: bool= True,decimal_sep: str='.', delimiter: str=';', timezone: str='UTC', aggregation: str='avg'):
        """Returns a csv file with the data of a given list of variables
        """
        # columns: field: "stream_id:sensorid" or sensorid
        # headers: ["temperature", "C"]
        
        substring = ''
        filename = ''
        streams = set()
        for var in variables:
            stream = self._get_stream_name_for_sid_vid(var.sid, var.id)
            streams.add(stream)
            s = f"""{{field: "{var.sid}:{var.index}.{aggregation}", 
            headers: ["{var.name}","{stream}","{aggregation}", "{var.unit or ""}"]}},"""
            substring += s
            
        filename += '_'.join(streams) + '_' + start + '_' + end + '_' + resolution + '_' + aggregation + '.csv'
        start = str(self.convert_datetime_to_unix(start))
        end = str(self.convert_datetime_to_unix(end))
        self.query = f"""
            {{
                exportCSV(
                    resolution: {resolution}
                    from: {start},
                    to: {end},
                    timezone: "{timezone}",
                    filename: "{filename}"
                    columns: [
                        {{
                            field: "ts",
                            headers: ["datetime"],
                            dateFormat: "%Y-%m-%dT%H:%M:%S"
                        }},
                        {{
                            field: "ts",
                            headers: ["time","","","[s since 01.01.1970]"]
                        }},
                        {substring}
                    ]
                ) {{
                    file
                }}
            }}
        """
        url_list = self.url+'/__api__/gql'
        headers = {'Authorization': 'Bearer ' + self.login_token["access_token"]}
        res = requests.post(url_list, json={'query':self.query}, headers = headers, stream=streaming)
        if res.status_code == 200 and not "errors" in res.text: 
            
            if write_file:
                csv_file = open(filepath + filename, 'wb')
                if stream == False:
                    content = res.content
                    csv_file.write(content)
                    csv_file.close()
                else:
                    for chunk in res.iter_content(chunk_size=1024):
                        csv_file.write(chunk)
                        csv_file.flush()
                    if return_df == False:
                        csv_file.close()
                        
            
            # return as df
            # TODO set columns to specific dtypes (more performance and pandas dynamic checking is buggy)
            if return_df == True and stream==False or write_file==False:
                return pd.read_csv(BytesIO(res.content), delimiter=delimiter)
            elif return_df == True:
                return pd.read_csv(filepath + filename, delimiter=delimiter)
        else:
            error = json.loads(res.text)
            logging.error(f"Fetching csv Data failed! \nResponse Code: {res.status_code} \nReason: {res.reason}\nMsg: {error['errors'][0]['message']}")
            
    
    def __get_column_names(self, sid:str, index_list:list): 
        """
        private helper method to get the column names for df
        Args:
            sid (str): Stream Id
            index_list (list): List of variable indices e.g. "a10"

        Returns:
            _type_: _description_
        """
        col_names = ["Time"]
        res = self.filter_var_attr("sid", sid)
        if res != None:
            for i in index_list: 
                for x in res:
                    if x.index == i:
                        col_names.append(x.name)
            return col_names
        return None 

    @staticmethod
    def convert_datetime_to_unix(datetime:str):
        """staticmethod for converting timestamps to UTC

        Args:
            datetime (str): required date format '%Y-%m-%d %H:%M:%S'

        Returns:
            int: UNIX timestamp
        """
        try: 
            date_time_obj = dt.datetime.strptime(datetime, '%Y-%m-%d %H:%M:%S')
            timestamp_utc = date_time_obj.replace(tzinfo=tz.gettz('UTC'))
            timestamp_local = timestamp_utc.astimezone(tz.gettz('Europe / Paris'))
            timestamp= int(dt.datetime.timestamp(timestamp_local)) * 1000 
            return timestamp
        except Exception as err:
            logging.error("Fehler beim konvertieren des Zeitstempels:", dt.datetime.now, err)

    def get_measurement_limit(self, sid:str, limit:str): 
        """
             'get measurement between time start and time stop'

        Args:
            sid (str): Stream Id
            limit (str): 
        """
        query_measurement="{\n  measurementPeriods(sid: \"" + sid + "\",from: 0,\n    to: 9999999999999, limit: "+str(limit)+", sort: DESC) \
        {\n    minTs\n    maxTs\n    mid\n    sampleRate\n  }\n}"
        url_list=self.url+'/__api__/gql'
        headers = {'Authorization': 'Bearer ' + self.login_token["access_token"]}
        try:
            res = requests.post(url_list, json = {'query':query_measurement}, headers = headers)
            if res.status_code == 200: 
                self.request_measurement_res = res.json()
            else: 
                logging.error(f"Fetching variable info failed! \nResponse Code:{res.status_code} \nReason: {res.reason}")
        except Exception as err:
            logging.warning(err)

    def print_measurement(self): # TODO test function
        limit = len(self.request_measurement_res['data']['measurementPeriods'])
        measurement_list = np.zeros((int(limit),2))
        for l in range(0,int(limit)):
        #print("start : ",request_measurement_res['data']['measurementPeriods'][l]['minTs'],"stop:",request_measurement_res['data']['measurementPeriods'][l]['maxTs'])
            measurement_list[l,0] = self.request_measurement_res['data']['measurementPeriods'][l]['minTs']
            measurement_list[l,1] = self.request_measurement_res['data']['measurementPeriods'][l]['maxTs']
        return(measurement_list)
        
    #### ubdf importer 
    def create_import_session_udbf(self, sid:str, stream_name:str):
        """ 
        method to import udbf file with http API

        Args:
            sid (str): _description_
            stream_name (str): _description_

        Returns:
            _type_: _description_
        """

        url_list = self.url + '/history/data/import'
        param = {"Type":"udbf",
                "SourceID": sid ,
                "SourceName": stream_name ,
                "MeasID":"",
                "SessionTimeoutSec":"300",
                "AddTimeSeries":"false",
                "SampleRate":"-1",
                "AutoCreateMetaData":"true"
                }
        headers = {'Content-Type':'application/json','Authorization': 'Bearer ' + self.login_token["access_token"]}
        try:
            res = requests.post(url_list,headers=headers,json=param)
            if res.status_code == 200: 
                self.import_session_res_udbf=res.json()
            else:   
                logging.error(f"Creating import session failed! \nResponse Code:{res.status_code} \nReason: {res.reason}")
        except Exception as err:
            logging.error(f"create_import_session failed:{err}")
            res="error"
        return (res)
    
    def import_file_udbf(self, file):
        """method to import udb file with http API

        Args:
            file (udbf): GI specific file format 

        Returns:
            http obj:  Server response 
        """

        self.session_ID = str(self.import_session_res_udbf['Data']['SessionID'])
        url_list = self.url +'/history/data/import/' + self.session_ID
        header_list = {'Content-Type':'application/octet-stream','Authorization': 'Bearer ' + self.login_token["access_token"]}
        try:
            res = requests.post(url_list,headers=header_list,data=file)
            if res.status_code == 200: 
                logging.info("ubdf succesfully imported")
            else:   
                logging.error(f"Import failed! \nResponse Code:{res.status_code} \nReason: {res.reason}")
        except Exception as err:
            logging.error(f"import udbf failed:{err}")
            res = "error"
        return (res)  
        
    #### csv importer 
    def create_import_session_csv(self, stream_ID:str, stream_Name:str, csv_config: CsvConfig, create_meta_data:bool = True, session_timeout:int = 60):
        """method to import csv file with http API

        Args:
            stream_ID (str): stream id
            stream_Name (str): buffername for new stream

        Returns:
            http obj:  Server response 
        """
        
        url_list = self.url+'/history/data/import'
        param = { "Type":"csv",
                "SourceID":stream_ID,
                "SourceName":stream_Name,
                "MeasID":"",
                "SessionTimeoutSec":str(session_timeout),
                "AddTimeSeries":"false",
                "SampleRate":"-1",
                "AutoCreateMetaData": str(create_meta_data).lower(),
                "CSVSettings": csv_config.get_config()
                }

        headers = {'Content-Type':'application/json','Authorization': 'Bearer ' + self.login_token["access_token"]}
        try:
            print(param)
            res = requests.post(url_list, headers = headers, json = param)
            if res.status_code == 200: 
                self.import_session_res_csv = res.json()
                self.import_session_csv_current = {'stream_id': stream_ID, 'ts': time.time(), 'timeout': session_timeout}
            else:   
                logging.err(f"Creating import session failed! \nResponse Code:{res.status_code} \nReason: {res.reason}")
        except Exception as e:
            logging.error(f"create_import_session failed:{e}")
            res = "error"
        return (res)

    def __import_file_csv(self, file):
        """method to import csv file with http API

        Args:
            file (csv): csv file for upload

        Returns:
            http obj:  Server response 
        """

        self.session_ID = str(self.import_session_res_csv['Data']['SessionID'])
        url_list = self.url + '/history/data/import/' + self.session_ID
        headers = {'Content-Type':'text/csv','Authorization': 'Bearer ' + self.login_token["access_token"]}
        try:
            res = requests.post(url_list, headers = headers, data = file)
        except Exception as e:
            logging.error("import csv failed ,code:{}".format(res.status_code))
            res = "error"
        return (res)

    def upload_csv_file(self, stream_name:str, file_path:str, py_formatter: str = None, csv_config: any = None): 
        """this method performs preparatory functions for csv import

        Args:
            stream_name (str): 
            file_path (str):
        """
        #**************************************
        #    read und check csv file
        #**************************************
        try:
            # py_formatter is delivered if python parses dates differently than c++ (backend parsing)
            # e.g. py_formatterClmn1 = "%d.%m.%Y %H:%M:%S.%f" for python, while DateTimeFmtColumn1: str = "%d.%m.%Y %H:%M:%S.%F" for API config
            if py_formatter == None:
                py_formatterClmn1 = self.DateTimeFmtColumn1
    
            first_lines = pd.read_csv (file_path, encoding = 'utf-8', nrows = 10, sep = ';') #decofing AINSI files on windows or linux
            read_date = first_lines.iat[self.ValuesStartRowIndex-1, 0] # we read the first measurement line, first coulmn
            read_date = Helpers.remove_hex_from_string(read_date) # rm cloud exported hex containments

            if (self.DateTimeFmtColumn2 == "" and self.DateTimeFmtColumn3 == ""):
                date_time_obj = dt.datetime.strptime(read_date+";",py_formatterClmn1+";"+self.DateTimeFmtColumn2)
            elif self.DateTimeFmtColumn2 != "":
                read_time=first_lines.iat[self.ValuesStartRowIndex-1,1]     # we read the first measurement lin, second coulmn 
                date_time_obj = dt.datetime.strptime(read_date+";"+read_time,py_formatterClmn1+";"+self.DateTimeFmtColumn2)
            
            #handle timestamp
            csv_timestamp_utc = date_time_obj.replace(tzinfo=tz.gettz('UTC'))
            csv_timestamp_local = csv_timestamp_utc.astimezone(tz.gettz('Europe / Paris'))
            csv_timestamp = dt.datetime.timestamp(csv_timestamp_local)
        except (FileNotFoundError) as e:
            logging.error('File path of csv to import is wrong.')
        except Exception as e:
            logging.error('Could not read the csv - check the config:', e)
            csv_timestamp=0

        timestamp_tmp = dt.datetime.fromtimestamp(csv_timestamp)
        timestamp_tmp.strftime('%Y-%m-%d %H:%M:%S')
        logging.info(f"First csv timestamp {csv_timestamp,timestamp_tmp.strftime('%d.%m.%Y %H:%M:%S')}")


        #**************************************
        #    check if stream exists
        #**************************************
        if stream_name in self.streams:
            write_ID = stream_name[stream_name].id
            reprise = 1# in this case reprise =1
            logging.info("Stream already existing in GI.Cloud import will be continued - {}".format(write_ID))
        else:
            write_ID=str(uuid.uuid4())
            reprise = 0# in this case reprise =1
            logging.info("Stream not existing in GI.Cloud import will be initialised - {}".format(write_ID))
            
        if (utils.is_valid_uuid(write_ID) == False):
            logging.error(f'uuid is wrong{write_ID}; break')
            return
        
        #**************************************
        #            check last imported timestamps
        #**************************************
        if reprise == 1:
            try:
                last_timestamp = self.streams[stream_name].last_ts
                timestamp_end_s = dt.datetime.utcfromtimestamp(last_timestamp/1000)
                print("last utc imported timestamp: {}".format((last_timestamp/1000),timestamp_end_s.strftime('%Y-%m-%d %H:%M:%S')))
            except:
                print("probably no stream existing")
        else:
            last_timestamp = 0
            print("stream_empty timestamp:".format(last_timestamp))

        #**************************************
        #           Upload file
        #**************************************
        if csv_config == None: 
            csv_config = self.csv_config
        if csv_timestamp > last_timestamp / 1000:
            # reuse import session if possible
            if not self.__import_session_valid(write_ID):
                self.create_import_session_csv(write_ID, stream_name, csv_config)
            with open(file_path, 'rb') as f:
                data_upload = f.read()

            response = self.__import_file_csv(data_upload)
            #time.sleep(5)

            if response.status_code == 200:
                logging.info(f"import of {file_path} was successful")
                return write_ID
            else:   
                logging.error(f"Import failed! \nResponse Code:{response.status_code} \nReason: {response.reason}")
                
            
            # free up resources 
            #self.delete_import()
        else:
            logging.error("Import failed : first imported csv value  is before the last database timestamp, but must begin after")
        return None
    
    def __import_session_valid(self, stream_id):
        if self.import_session_csv_current == None:
            return False
        elif self.import_session_csv_current['stream_id'] == stream_id and \
            (self.import_session_csv_current['ts'] - time.time()) +5 < self.import_session_csv_current['timeout']:
            return True
        
    def delete_import_session(self):
        """
         method to delete session http API

        Returns:
             http obj:  Server response 
        """
       
        url_list = self.url +'/history/data/import/' + self.session_ID
        headers = {'Authorization': 'Bearer ' + self.login_token["access_token"]}
        try:
            res = requests.delete(url_list, headers = headers)
            if res.status_code == 200: 
                logging.info("Import session is closed")
            else:   
                logging.error(f"Failed closing import session! \nResponse Code:{res.status_code} \nReason: {res.reason}")
        except Exception as e:
            logging.error(f"delete failed: {e}")
            res="error"
        return (res)

    ################# Read and Write Single Values out of live datastreams
    ############### notice postprocessed data do not work with this functions 

    def read_value(self, var_ids:list): 
        """read online value, enter a list of variable IDs
        Args:
            var_ids (list): #For example var_ids=["47f32894-c6a0-11ea-81a1-02420a000368"]

        Returns:
            list: current live value
        """

        url_list = self.url + '/online/data'
        param = {"Variables":var_ids,"Function":"read"}
        headers = {'Content-Type':'application/json','Authorization': 'Bearer ' + self.login_token["access_token"]}
        try:
            res = requests.post(url_list, headers = headers, json = param)
            if res.status_code == 200: 
                current_live_value = res.json()
                logging.info(f"Current live Value: {current_live_value}")
                return current_live_value
            else:
                logging.error(f"failed! \nResponse Code:{res.status_code} \nReason: {res.reason}") 
        except Exception as e:
            logging.error(e)
        return None

    def write_value_on_channel(self, var_ids:list, write_list:list):
        """ read online value, enter a list of variable IDs
        Args:
            var_ids (list): #For example var_ids=["47f32894-c6a0-11ea-81a1-02420a000368"]
            write_list (list): #For example write_list=["0"]

        Returns:
            http obj:  Server response 
        """

        url_list = self.url + '/online/data'
        param = {"Variables":var_ids, "Values": write_list ,"Function":"write"}
        headers = {'Content-Type':'application/json','Authorization': 'Bearer ' + self.login_token["access_token"]}
        try:
            res = requests.post(url_list, headers = headers, json = param)
            if res.status_code == 200: 
                write_value_res=res.json() 
                logging.info("Data Successfully written")
                return write_value_res
            else:
                logging.error(f"failed! \nResponse Code:{res.status_code} \nReason: {res.reason}") 
        except Exception as e:
            logging.error(e)
        return None
    
    def get_gistreamvariables(self, stream: str, variables: list) -> list:
        """Give GIStreamVariable for according stream and variable name"""
        gi_vars = []
        for var in variables:
            gi_vars.append(list(self.find_var(stream + '__' + var).values())[0])
        return gi_vars

class Helpers():
    @staticmethod 
    def remove_hex_from_string(str):
        """Remove hex value from input string"""
        return re.sub(r'[^\x00-\x7f]',r'', str)

class Resolution(Enum):
    MONTH = 'MONTH'
    WEEK = 'WEEK'
    DAY = 'DAY'
    HOUR = 'HOUR'
    QUARTER_HOUR = 'QUARTER_HOUR'
    MINUTE = 'MINUTE'
    SECOND = 'SECOND'
    HZ10 = 'HZ10'
    HZ100 = 'HZ100'
    KHZ = 'KHZ'
    KHZ10 = 'KHZ10'
    NANOS = 'nanos'

@dataclass()
class GIStream:
    '''Object for tracking available streams''' 
    name: str
    id: str
    sample_rate_hz: str
    first_ts: int
    last_ts: int
    index: int





# -*- coding: utf-8 -*-
"""
Created on Wed Apr  8 16:47:07 2020

@author: jouanb
"""
from ctypes import*
import ctypes
import os


#//////////////////////////////////////////////////////////////////////////////////////////
#/*------------- PostProcess buffer server handling -------------------------------------*/
#/*																						*/
#/*	Description:																		*/
#/*																						*/
#/*		Following functions allow creation of PostProcess buffers / data stream			*/
#/*		Depending on environmental settings, different data backends are supported   	*/
#/*																						*/
#//////////////////////////////////////////////////////////////////////////////////////////
#/**
# * @brief Create new PostProcess buffer / SystemStream
# *
# * @param sourceID			source UUID (SID) of this buffer
# * @param sourceName		name of this buffer
# * @param measurementID		measurement UUID (MID) of the actual mesurement
# * @param measurementName	name of the actual measurement
# * @param sampleRateHz		the desired sample rate for this measurement
# * @param bufferSizeByte	the maximum size of this buffer in bytes
# * @param segmentSizeByte	the size of a buffer segment (if supported)
# * @param retentionTimeSec  data retention time of this buffer (if supported)
# *
# * @param bufferHandle		the result handle
# * @param errorMsg			buffer for error message text if not successful
# * @param errorMsgLen		length of the error message buffer
# *
# * @return General return codes
# */

#const char* -->c_char_p
#double      -->c_double
#double*      -->POINTER(c_double)
#uint64_t    -->? probably c_int
#int32_t      -->c_int
#int32_t*    -->POINTER(c_int)
#char*       -->c_char_p
#uint32_t    -->c_int


class PostProcessBufferServer():
    def __init__(self):

        try:
            #chemin=os.path.abspath(".\\Giutility\\giutility_x64.dll")
            chemin = "libGInsUtility.so"
            
            self.GINSDll = ctypes.cdll.LoadLibrary(chemin)
            
            
            print("64 bit DLL imported")
            print(chemin)
        except OSError as err:
            print("err: {}\ncan not imported libGInsUtility.so", format(err))

        #New v3.0
        self.GINSDll._CD_eGateHighSpeedPort_PostProcessBufferServer_Create.argtypes = [c_char_p,c_char_p,c_double,c_ulonglong,c_ulonglong,c_double,c_char_p,POINTER(c_int),c_char_p,c_int]
        self.GINSDll._CD_eGateHighSpeedPort_PostProcessBufferServer_AddVariableDefinition.argtypes = [c_int,c_char_p,c_char_p,c_char_p,c_int,c_int,c_int,c_int,c_double,c_double,c_char_p,c_int]        
        self.GINSDll._CD_eGateHighSpeedPort_PostProcessBufferServer_Initialize.argtypes =[c_int,c_int,c_char_p,c_int]
        self.GINSDll._CD_eGateHighSpeedPort_PostProcessBufferServer_WriteDoubleToFrameBuffer.argtypes =[c_int,c_int,c_int,c_double,c_char_p,c_int]
        self.GINSDll._CD_eGateHighSpeedPort_PostProcessBufferServer_WriteTimestampToFrameBuffer.argtypes =[c_int,c_int,c_ulonglong,c_char_p,c_int]
        self.GINSDll._CD_eGateHighSpeedPort_PostProcessBufferServer_AppendFrameBuffer.argtypes =[c_int,c_char_p,c_int]
        self.GINSDll._CD_eGateHighSpeedPort_PostProcessBufferServer_AppendDataBuffer.argtypes =[c_int,c_char_p,c_ulonglong,c_char_p,c_int]
        self.GINSDll._CD_eGateHighSpeedPort_PostProcessBufferServer_Close.argtypes =[c_int,c_char_p,c_int]
        self.GINSDll._CD_eGateHighSpeedPort_SleepMS.argtypes =[c_int]
        #self.GINSDll._CD_eGateHighSpeedPort_PostProcessBufferServer_CreateUDBFFileBuffer.argtypes = [c_char_p,c_char_p,c_char_p,c_double,c_ulonglong, c_int16,POINTER(c_int32),c_char_p,c_int32]
        self.buffersize=50000000
        self.segmentsize=50000000
        self.bufferHandle=c_int(-1)
        self.errorlen=30
        self.frameBufferLength=10
        self.errorMsg=ctypes.create_string_buffer(30)
        

    def create_UDBF_file_buffer(self, file_path, stream_id, stream_name, sample_rate):
        self.max_length_in_sec = 3600 #the maximum length of a file in seconds
        self.options = 0 # bitset for future options- what does that mean???

        self.ID = stream_id.encode('UTF-8')
        self.BufferName = stream_name.encode('UTF-8')
        self.StreamSampleRate = sample_rate

        #TODO when udbf upload is rdy - add function here

    def create_buffer(self,ID,BufferName,StreamSampleRate):
        """create buffer"""
        self.ID= ID.encode('UTF-8')
        self.BufferName=BufferName.encode('UTF-8')
        self.StreamSampleRate=StreamSampleRate
        
        dataTypeIdent='raw'
        self.dataTypeIdent=dataTypeIdent.encode('UTF-8')
        #self._CD_eGateHighSpeedPort_PostProcessBufferServer_Create.argtypes = [c_char_p,c_char_p,c_double,c_int,c_int,c_double,c_char_p,POINTER(c_int),c_char_p,c_int]
        ret=self.GINSDll._CD_eGateHighSpeedPort_PostProcessBufferServer_Create(self.ID, self.BufferName, self.StreamSampleRate, self.buffersize, self.segmentsize,0, self.dataTypeIdent, byref(self.bufferHandle),self.errorMsg, self.errorlen)
        if ret!=0:
            print("Error at create buffer!")

    def add_channel(self,variableID,variableName,Unit):
        """add channels"""
        self.variableID=variableID.encode('UTF-8')
        self.variableName=variableName.encode('UTF-8')
        self.Unit=Unit.encode('latin-1')
        self.DataTypeCode=8
        self.VariableKindCode=6
        self.Precision=4
        self.FieldLength=8
        self.RangeMin=-100
        self.RangeMax=100
        self.errorMsgLen=self.errorlen
        #self._CD_eGateHighSpeedPort_PostProcessBufferServer_Create.argtypes = [c_char_p,c_char_p,c_double,c_int,c_int,c_double,c_char_p,POINTER(c_int),c_char_p,c_int]
        ret=self.GINSDll._CD_eGateHighSpeedPort_PostProcessBufferServer_AddVariableDefinition(self.bufferHandle, self.variableID, self.variableName, self.Unit,self.DataTypeCode,self.VariableKindCode,self.Precision,self.FieldLength,self.RangeMin,self.RangeMax,self.errorMsg,self.errorMsgLen)
        if ret!=0:
            print("Error adding channel")
            
    def init_buffer(self):
        """Init buffer"""
        ret=self.GINSDll._CD_eGateHighSpeedPort_PostProcessBufferServer_Initialize(self.bufferHandle, self.frameBufferLength, self.errorMsg, self.errorMsgLen)
        if ret!=0:
            print("Error buffer initialized")
            self.GINSDll._CD_eGateHighSpeedPort_PostProcessBufferServer_Close(self.bufferHandle,self.errorMsg, self.errorMsgLen)
        else :
            print("success buffer ini",self.bufferHandle) 
    def write_timestamps(self,frameIndex,timestamp_ns):
        #self.frameIndex=frameIndex
        #self.valueInt=timestamp_ns
        ret=self.GINSDll._CD_eGateHighSpeedPort_PostProcessBufferServer_WriteTimestampToFrameBuffer(self.bufferHandle,frameIndex,timestamp_ns,None, self.errorMsgLen)
        if ret!=0:
            print("error writing timestamps")
        return (ret)
        #else :
        #    print("success writing timestamps","buffer",self.bufferHandle,"frameindex",frameIndex,"ts",timestamp_ns,)
            
    def write_to_framebuffer(self,frameIndex,variableIndex,valueDouble):
        #self.frameIndex=frameIndex
        #self.variableIndex=variableIndex
        #self.valueDouble=valueDouble
        ret=self.GINSDll._CD_eGateHighSpeedPort_PostProcessBufferServer_WriteDoubleToFrameBuffer(self.bufferHandle,frameIndex,variableIndex,valueDouble,self.errorMsg, self.errorMsgLen)
        if ret!=0:
            print("error filling frame")
        return (ret)
        #else :
        #    print("success filling fram ret",ret)
            #print("error message",self.errorMsg)

    def append_to_framebuffer(self):
        ret=self.GINSDll._CD_eGateHighSpeedPort_PostProcessBufferServer_AppendFrameBuffer(self.bufferHandle,self.errorMsg,self.errorMsgLen)
        if ret!=0:
            print("error append to framebuffer - code:" + str(ret))
            print(self.errorMsg.value, "len:",self.errorMsgLen )
            
    def append_data_to_framebuffer(self,data,dataLength):
        self.data=data
        self.dataLength=dataLength
        ret=self.GINSDll._CD_eGateHighSpeedPort_PostProcessBufferServer_AppendDataBuffer(self.bufferHandle,self.data,self.dataLength,self.errorMsg,self.errorMsgLen)
        if ret!=0:
            print("error append_data")

    def close_buffer(self,bufferHandle):
        self.bufferHandle=bufferHandle
        ret=self.GINSDll._CD_eGateHighSpeedPort_PostProcessBufferServer_Close(self.bufferHandle,self.errorMsg, self.errorMsgLen)
        if ret!=0:
            print("error closing connection")
            
    def buffer_sleep(self,sleep_time):
        #self.sleep_time=sleep_time
        ret=self.GINSDll._CD_eGateHighSpeedPort_SleepMS(sleep_time)
        if ret!=0:
            print("error by sleeping")




from datetime import datetime 
import pytz
import pandas as pd

def append_timezone_ts_to_df(df, ts_col, new_timezone, old_timezone='Africa/Accra'):
    """Appends a column of to_new timezone converted timestamps to given Dataframe.
        'Africa/Accra' -> UTC/GMT +0
        America/Phoenix -> ts of arizona
    Args:
        df (DataFrame): 
        ts_col (String): Column name
        old_timezone ( String): Timezone of current ts
        new_timezone (String): Timezone of destination ts

    Returns:
        [DataFrame]:
    """
    
    
    # Convert strings into pytz objects
    #old_timezone = pytz.timezone(old_timezone)
    #new_timezone = pytz.timezone(new_timezone)
    
    #df = df.assign(ts_new = lambda x: (old_timezone.localize(x[ts_col]).astimezone(new_timezone)))#
    

    df['ts_c'] = pd.to_datetime(df[ts_col]/1000, unit='s')
    
    df.ts_c = df.ts_c.dt.tz_localize('UTC').dt.tz_convert(new_timezone)

    return df

def replace_ts_with_timezone_ts(df, ts_col, new_timezone, old_timezone='Africa/Accra'):
    
    df[ts_col] = pd.to_datetime(df[ts_col]/1000, unit='s')
    
    df[ts_col] = df[ts_col].dt.tz_localize('UTC').dt.tz_convert(new_timezone)
    
    return df



def ts_to_dateobj_with_timezone(ts, timezone):
    """
    Generate a datetime object with corresponding timezone from given timestamp
    As unix timestamps are always UTC (GMT+0)
    Args:
        ts ([float]): Timestamp
        timezone (String): target Timezone

    Returns:
        [datetime.datetime]: datetime object
    """
    # UTC to not take localtime into account
    date_obj = datetime.utcfromtimestamp(ts/1000) # Convert into datetime object
    
    date_obj = date_obj.astimezone(pytz.timezone(timezone)) # Apply Arizona timezone
    
    return date_obj


def aggregate_df_as_resample(df, aggregation, math_operation, date_column = '',):
    """
    Aggregates data together by a given aggregation type e.g. ('H'(Hour) 'D'(Day))
    and uses mathematical operation on them like sum(), mean(), max()
    adds column 'datetime'
    

    Args:
        df ([DataFrame]): [description]
        aggregation ([type]): [description]
        math_operation ([String]): [description]

    Returns:
        [DataFrame]: [copied and changed dataframe]
    """
    if date_column == '':
        df['datetime'] = pd.to_datetime(df['ts']/1000, unit='s')
        resample = df.set_index('datetime')
    else:
        resample = df.set_index(date_column)
        
    if math_operation == 'sum':
        resample = resample.resample(aggregation).sum()
    elif math_operation == 'max':
        resample = resample.resample(aggregation).max()
    elif math_operation == 'mean':
        resample = resample.resample(aggregation).mean()

    # Reset to normal indexed DF
    resample = resample.reset_index()

    return resample

def bin_df_column_values(df, column, ratio):
    """
    Binning values together e.g. Tamd:ratio5 (23 -> 25)

    Args:
        df (DataFrame): given
        column (var): column of DF
        ratio (double): binning Ratio

    Returns:
        DataFrame: 
    """
    df[column] = pd.to_numeric(round(df[column]/ratio + 0.5) * ratio)
    
    return df

def drop_n_rows(df: pd.DataFrame, N: int) -> pd.DataFrame:
    return df.iloc[N:, :]

def append_hod_yymmdd(df: pd.DataFrame, ts: str, timezone: str = 'America/Phoenix') -> pd.DataFrame:
    """Based on timestamp column and timezone, add corresponding yymmdd and hhmm (hour and minutes) to df as columns"""
    yymmdd, yymm, hhmm = [],[],[]
    
    for indx, ts in enumerate(df[ts]):
        date_obj = datetime.fromtimestamp(ts)
        date_obj = date_obj.astimezone(pytz.timezone(timezone))
        yymmdd.append(date_obj.strftime('%Y%m%d')[2:])
        yymm.append(date_obj.strftime('%Y%m')[2:])
        hhmm.append(int(date_obj.strftime('%H%M')))
    df['YYMM'] = yymm
    df['YYMMDD'] = yymmdd
    df['HHMM'] = hhmm
    return df

def datetime_column_from_ts(df: pd.DataFrame, ts: str) -> pd.DataFrame:
    """Change timestamp column into column filled with datetime objects"""
    df['date'] = pd.to_datetime(df[ts], errors='coerce', unit='s')
    return df


# This module implements domain functions of the python write section


import numpy as np
import pandas as pd
from datetime import datetime
import datetime as dt

from pandas.core.arrays.sparse import dtype


def calc_sums(arr, frequenzy, aggregation='D', ts_index=0):
    """Calculates sum of given measured frequenzy (Hz) and aggregation type(e.g. Hour)

    Args:
        arr ([type]): Array
        ts_offset ([type]): array index to skip
        frequenzy ([type]): (Hz) given as 's' 'm' ...
        aggregation ([type]): (Hour, Day, Month)
    """


    # Convert array into dataframe for sum operations
    df = pd.DataFrame(data=arr)
    
    if frequenzy == 's' or frequenzy == 'm':
        df['datetime'] = pd.to_datetime(df[0]/1000, unit='s') #unit=(D,s,ms,us,ns)
    
    resample = df.set_index('datetime')
    resample = resample.resample(aggregation).sum()

    resample = resample.reset_index()

    # Fix timestamps (got summed up before)
    for index, row in resample.iterrows():
        resample[ts_index][index] = datetime.timestamp(resample['datetime'][index]) * 1000 

    # Drop datetime column to get original array shape
    resample = resample.drop(['datetime'], axis=1)

    resample = resample.to_numpy()
    return resample

        


from iniconfig import ParseError
import sqlalchemy
from sqlalchemy import exc
import logging
import pandas as pd

import pymysql
pymysql.install_as_MySQLdb()

from gimodules.cloudconnect.utils import split_dates, get_dates_from_string, replace_dates_in_string
# Set output level to INFO because default is WARNNG
logging.getLogger().setLevel(logging.INFO)

class MySQLConnect():
   
   def __init__(self, db_host: str, db_user: str, db_port: str, db_pass: str, db_name: str) -> None:
      self.connect_to_db(db_host, db_user, db_pass, db_port, db_name)
      
   
   def connect_to_db(self, db_host, db_user, db_pass, db_port, db_name):

      # Convert utf8mb4 so mysql < 5.0.2
      connect_string = 'mysql+pymysql://{}:{}@{}:{}/{}?charset=utf8'.format(db_user, db_pass, db_host, db_port, db_name)
      
      try:
         engine = sqlalchemy.create_engine(connect_string, echo=True)

         self.connection = engine.connect()

         # Check wether conn is none or not to determine if connection is successful
         return self.connection
      except Exception as e:
         # call method to visualize failed connection
         print('Connection to ', db_host, ' could not be established.', e) 
   
   def query_sql(self, query: str, set_columns: bool = True, *args, **kwargs) -> pd.DataFrame:
      
      trans = self.connection.begin()
      
      try:
         res = self.connection.execute(query)
         
         trans.commit()
         ref_data = res.fetchall()
         df = pd.DataFrame(ref_data)
         #if set_columns:
            #df.columns = res.keys()
         return df
         
      except exc.SQLAlchemyError:
         logging.info('SQL Query is wrong.')



   def multiple_calls(self, query: str, N: int, *args, **kwargs) -> pd.DataFrame:
      """Multiple query call to avoid timeout error.
      Parse the start and end date, split it into multiple calls
      and concatenate results.
      """
      dates = get_dates_from_string(query)
      if len(dates) != 2:
         raise ParseError('Could not parse out the dates of query - lookup format!')

      # e.g. if N = 2 -> list size is 3
      dates_to_query = split_dates(dates[0], dates[1], N)
      
      # 
      frames = []
      i = 1
      while i < len(dates_to_query):
         temp_query = replace_dates_in_string(query, [dates_to_query[i-1], dates_to_query[i]])
         logging.info(f'Querying from {dates_to_query[i-1]} to {dates_to_query[i]}')
         
         frames.append(self.query_sql(temp_query))
         i += 1
      
      return pd.concat(frames, ignore_index=True)
   


import logging
import os
import re
import zipfile

logging.getLogger().setLevel(logging.INFO)

class BaseCsv():
   """
   Handles functions that help working with multiple coherent csv files
   """

   @staticmethod
   def list_all_files_inside_dir(path):
      onlyfiles = [f for f in os.listdir(path) if os.path.isfile(os.path.join(path, f))]
      sortet_files = sorted(onlyfiles)
      return sortet_files

   @staticmethod
   def list_all_regex_files_inside_dir(path, regex):
      return [f for f in os.listdir(path) if re.match(r'{}'.format(regex), f)]

   @staticmethod
   def sort_list(list):
      """ Sort list on splitting with first '-' and '.' """
      return sorted(list, key=lambda x: int(x.split('-')[1].split('.')[0]))
   
   @staticmethod
   def sort_list_on_splitter(list, split1, split2):
      """ Sort list on splitting with first '-' and '.' """
      return sorted(list, key=lambda x: int(x.split(split1)[1].split(split2)[0]))
      
   @staticmethod
   def split(filehandler, delimiter=',', row_limit=10000,
          output_name_template='%s.csv', output_path='.', keep_headers=True):
      """
      split(open('/your/path/input.csv', 'r'));

      Args:
          filehandler (_type_): _description_
          delimiter (str, optional): _description_. Defaults to ','.
          row_limit (int, optional): _description_. Defaults to 10000.
          output_name_template (str, optional): _description_. Defaults to 'output_%s.csv'.
          output_path (str, optional): _description_. Defaults to '.'.
          keep_headers (bool, optional): _description_. Defaults to True.
      """
      import csv
      reader = csv.reader(filehandler, delimiter=delimiter)
      current_piece = 1
      current_out_path = os.path.join(
         output_path,
         output_name_template % current_piece
      )
      current_out_writer = csv.writer(open(current_out_path, 'w'), delimiter=delimiter)
      current_limit = row_limit
      if keep_headers:
         headers = next(reader)
         current_out_writer.writerow(headers)
      for i, row in enumerate(reader):
         if i + 1 > current_limit:
               current_piece += 1
               current_limit = row_limit * current_piece
               current_out_path = os.path.join(
                  output_path,
                  output_name_template % current_piece
               )
               current_out_writer = csv.writer(open(current_out_path, 'w'), delimiter=delimiter)
               if keep_headers:
                  current_out_writer.writerow(headers)
         current_out_writer.writerow(row)
   
   @staticmethod
   def extract_zip(zip_file, targer_dir): 
      """
      Attention: path should not contain "work"
      """
      try: 
         with zipfile.ZipFile(zip_file, "r", compression='ZIP_STORED', allowZip64=True) as zip_ref: 
               zip_ref.extractall(targer_dir)
      except OSError as err: 
         print(err)
         logging.warning("ZipFile start \n {} \n Zipfile end".format(err))
         
   


import pytest
"""Placeholder for testing"""

def capital_case(x):
   if not isinstance(x, str):
        raise TypeError('Please provide a string argument')
   return x.capitalize()
   
def test_capital_case():
    assert capital_case('semaphore') == 'Semaphore'

def test_raises_exception_on_non_string_arguments():
    with pytest.raises(TypeError):
        capital_case(9)
        
def test_check_action():
    assert (2+2) == 4



import ipywidgets as widgets
from numpy import select  # import widgets for better user interaction
import datetime as dt
import numpy as np

# Letting the user choose a meas-period
"""
wid_meas=widgets.Dropdown(
    options=mess_start,
    value=mess_start[0],
    description='Measurement:',
    disabled=False,
)


stream_index = widgets.IntText(description='index:',value=0)
stream_ts_start = widgets.IntText(description='ts start:',value=measurement_list[0,0])
stream_ts_stop = widgets.IntText(description='ts stop:',value=measurement_list[len(measurement_list)-1,1])

wid_meas_end = widgets.Text(
    value=mess_end[0],
    placeholder='Paste ticket description here!',
    description='end:',
    #layout = widgets.Layout(width='400px'),
    disabled=False,
)

def meas_select_change(change):
    if change['type'] == 'change' and change['name'] == 'value':
        #print(str(wid_stream.value))
        stream_index.value=mess_start.index(wid_meas.value)
        wid_meas_end.value=mess_end[mess_start.index(wid_meas.value)]
        stream_ts_start.value=measurement_list[stream_index.value,0]
        stream_ts_stop.value=measurement_list[len(measurement_list)-1,1]
        display(stream_index,wid_meas,wid_meas_end,stream_ts_start,stream_ts_stop)
wid_meas.observe(meas_select_change,names='value')

display(wid_meas,wid_meas_end,stream_index,stream_ts_start,stream_ts_stop)

"""


class InputHandler():
    def __init__(self) -> None:
        pass

    def in_calc_measurements(self, sel_indices, channels):  # TODO rename function
        calculations = ("*", "/", "+", "-")
        temp = []
        self.sel_indices_map = {}
        counter = 0

        # Extract selected channel names
        for i in range(len(channels)):
            for j in range(len(sel_indices)):
                if (i == sel_indices[j]):
                    self.sel_indices_map.update({sel_indices[j]: counter})
                    counter += 1
                    temp.append(channels[i])

        self.sel_channel_1_widgets = [widgets.Dropdown(
            options=temp,
            description='Param_1:',
            disabled=False
        ) for chan in sel_indices]

        self.sel_calc_widgets = [widgets.Dropdown(
            options=calculations,
            description="Expression:",
            diabled=False
        ) for chan in sel_indices]

        self.sel_calc_2_widgets = [widgets.Dropdown(
            options=calculations,
            description="Expression:",
            diabled=False
        ) for chan in sel_indices]

        self.sel_param_widgets = [widgets.BoundedFloatText(
            value=1,
            description="Constant:",
            diabled=False
        ) for chan in sel_indices]

        temp2 = temp.copy()
        # As default, for ability to only use the constant
        temp2.insert(0, 'None')
        self.sel_channel_2_widgets = [widgets.Dropdown(
            options=temp2,
            description="Param_2:",
            disabled=False
        )for chan in sel_indices]

        self.sel_unit_widgets = [widgets.Text(
            description="Unit:",
            disabled=False
        ) for chan in sel_indices]

        # TODO place side by side
        # Display all widgets
        for i in range(len(self.sel_channel_1_widgets)):
            self.sel_channel_1_widgets[i].value = temp[i]
            display(self.sel_param_widgets[i], self.sel_calc_widgets[i], self.sel_channel_1_widgets[i],
                    self.sel_calc_2_widgets[i], self.sel_channel_2_widgets[i], self.sel_unit_widgets[i])

    def on_param_change(change):
        print()

    def in_calc_measurements_change(change):
        print()
        # if change['type'] == 'change' and change['name'] == 'value':

    def get_expr_param_for_sel_indices(self, sel_indices):
        if (not hasattr(self, "expr_param_map")):
            self.expr_param_map = {}
            """
            Expression map example: D = {'0': {'exp_c': '/', 'constant': '10', 'exp_2': '*' 'param_2': 'Gi_pyr'}} - 0 = channel/param_1
            """
        for i in range(len(sel_indices)):  # Not validated : could validate through map first
            self.expr_param_map.update(
                {i: {'param_1':self.sel_channel_1_widgets[i].value, 'exp_c': self.sel_calc_widgets[i].value, 'exp_2': self.sel_calc_2_widgets[i].value, 'constant': self.sel_param_widgets[i].value, 'param_2': self.sel_channel_2_widgets[i].value}})

        return self.expr_param_map
    
    def get_channel_name_by_index(widget, index):
        return widget.options[index]
    
    def get_channel_index_for_data(self, map_index, param, channel):
        if (channel == 1):
            return self.sel_channel_1_widgets[map_index].options.index(self.expr_param_map[map_index][param])
        elif (channel == 2):
            return self.sel_channel_2_widgets[map_index].options.index(self.expr_param_map[map_index][param])
    
    def declare_custom_units(self, sel_indices, unit_list):
        for i in range(len(sel_indices)):
            unit_list[sel_indices[i]] = self.sel_unit_widgets[i].value

    # above is mostly for python write
    #####################################################################
    def display_stream_channel_sel(self, conn_cloud, default_stream=0):
        """Displays different streams and lets user to pick different channels

        Args:
            conn_cloud (CloudRequest object): handles cloud requests (GQL requests)
            default_stream (int, optional): [description]. Defaults to 0.
        """
        if (default_stream != 0):
            default_stream = self.stream_list.index(default_stream)
        
        wid_stream=widgets.Dropdown(
            options=conn_cloud.stream_list,
            value=conn_cloud.stream_list[default_stream],
            description='Stream:',
            disabled=False,
        )

        stream_index = widgets.IntText()

        stream_ID_wid = widgets.Text(
            value=conn_cloud.stream_ID[default_stream],
            placeholder='Paste ticket description here!',
            description='Stream ID:',
            layout = widgets.Layout(width='400px'),
            disabled=True,
        )

        #map sources
        res=conn_cloud.variable_mapping(stream_ID_wid.value)
        index_var,name_var,unit_var,id_var=conn_cloud.get_var_mapping(conn_cloud.request_map_res)
        self.sel_stream_id = stream_ID_wid.value

        # We display interactively map sources
        self.wid_channel=widgets.SelectMultiple(
            options=name_var,
            #value=[name_var[3],name_var[8]], #preselected channels
            description='Channel:',
            layout = widgets.Layout(height='400px',width='400px'),
            disabled=False,
        )

        def dropdown_slect_change(change):
            if change['type'] == 'change' and change['name'] == 'value':
                self.wid_channel.disabled = True
                
                stream_index.value=conn_cloud.stream_list.index(wid_stream.value)
                stream_ID_wid.value=conn_cloud.stream_ID[stream_index.value]
                self.sel_stream_id = stream_ID_wid.value

                #we update the channels informations
                res=conn_cloud.variable_mapping(stream_ID_wid.value)
                index_var,name_var,unit_var,id_var=conn_cloud.get_var_mapping(conn_cloud.request_map_res)
                self.wid_channel.options=name_var
                #selected_index=[index_var[name_var.index(self.wid_channel.value[i])] for i in range(0,len(self.wid_channel.value))]
                #display the updated information selected_index=[index_var[name_var.index(wid_channel.value[i])] for i in range(0,len(wid_channel.value))]              
                self.wid_channel.disabled = False
                display(stream_index,conn_cloud.stream_IDD,self.wid_channel)
                
    
        wid_stream.observe(dropdown_slect_change,names='value')

        print("Choose the Stream and Variables of which you want to fetch data. \nPress CTRL to select multiple Variables.")
        display(wid_stream,stream_ID_wid,self.wid_channel)

    def get_selected_channel_indices(self, index_var, name_var):
        selected_index=[index_var[name_var.index(self.wid_channel.value[i])] for i in range(0,len(self.wid_channel.value))]
        return selected_index

    def display_date_pickers(self):
        """Display datepickers for data querying. Calculate timestamps on change of widget
        """
        # Create last week as default values
        today=dt.datetime.today()
        yesterday=dt.datetime.today() - dt.timedelta(days=1)
        a_week_ago=dt.datetime.today() - dt.timedelta(days=7)

        start_date_pick=widgets.DatePicker(
            value=a_week_ago,
            description='Start Date',
            disabled=False
        )

        end_date_pick=widgets.DatePicker(
            value=today,
            description='End Date',
            disabled=False
        )

        start_time_pick=widgets.Text(
            value='00:00:00',
            placeholder='Start time',
            description='String:',
            disabled=False
        )

        end_time_pick=widgets.Text(
            value='23:59:59',
            placeholder='Endtime',
            description='String:',
            disabled=False
        )
        
        

        def timedate_to_ts():
            time_start_obj = dt.datetime.strptime(start_time_pick.value,'%H:%M:%S').time()
            time_end_obj = dt.datetime.strptime(end_time_pick.value,'%H:%M:%S').time()
            start_day=dt.datetime.combine(start_date_pick.value,time_start_obj)
            end_day=dt.datetime.combine(end_date_pick.value,time_end_obj)
            self.timestamp_start = str(int(np.multiply(dt.datetime.timestamp(start_day),1000)))
            self.timestamp_stop = str(int(np.multiply(dt.datetime.timestamp(end_day),1000)))
            #print(self.timestamp_start + ' ' + self.timestamp_stop)

        

        out = widgets.Output() # To enable printing in same widget block
        display(out)
        
        @out.capture()
        def on_datetime_change(change):
            """On time or date change calculate these into new start/end timestamps

            Args:
                change (date/time is changed): 
            """
            if change['type'] == 'change' and change['name'] == 'value':
                timedate_to_ts()
        
        start_date_pick.observe(on_datetime_change, names='value')
        end_date_pick.observe(on_datetime_change, names='value')
        start_time_pick.observe(on_datetime_change, names='value')
        end_time_pick.observe(on_datetime_change, names='value')
        timedate_to_ts()
        display(start_date_pick,start_time_pick,end_date_pick,end_time_pick)

    def display_res_sel(self):
        self.wid_reso=widgets.Select(
            options=['MONTH','WEEK','DAY','HOUR','QUARTER_HOUR','MINUTE','SECOND','HZ10','HZ100','KHZ','KHZ10','nanos'],
            value='MINUTE',
            # rows=10,
            description='Resolution',
            disabled=False
        )
        display(self.wid_reso)

        self.sel_resolution = self.wid_reso.value

        def on_res_change(change):
            if change['type'] == 'change' and change['name'] == 'value':
                print(change)
                self.sel_resolution = change.new
        
        self.wid_reso.observe(on_res_change, names='value')


# Module to reuse predifened plots
# TODO - implement savefig's with correct path

from ipywidgets.widgets.widget_string import Label
import seaborn as sns
import matplotlib
import matplotlib.pyplot as plt # to plot graph
import matplotlib.dates as md
import pandas as pd # dataframe library
import numpy as np
import datetime as dt
import os
import math



# From guid 22
def plt_violin_seaborn(df, xval, yval, title, ylabel, figsize=None, start_yymmdd=None, end_yymmdd=None, start_yymm=None, end_yymm=None, *argv, **kwargs):

    #Meta information
    for key, value in kwargs.items():
        print(key + ' : ' + value)
    
    # args for costum description
    for arg in argv:
        print(arg)
    

    font = {
        'weight' : 'bold',
        'size'   : 25}

    matplotlib.rc('font', **font)
    if not start_yymmdd and not end_yymmdd:
        #print('From: ',start_date_pick.value, 'to: ', end_date_pick.value)
        print()
    else:
        print('From: ',start_yymmdd, 'to: ', end_yymmdd)

    if figsize:
        plt.figure(figsize=figsize)
    else:
        plt.figure()

    #Filter input date
    if start_yymmdd and end_yymmdd:
        df = df[df['YYMMDD'] <= int(end_yymmdd)]
        df = df[df['YYMMDD'] >= int(start_yymmdd)]
    if start_yymm and end_yymm:
        df = df[df['YYMM'] <= int(end_yymm)]
        df = df[df['YYMM'] >= int(start_yymm)]
        
        
    # https://www.graphpad.com/support/faq/violin-plots-and-logarithmic-axes/
    # Issue regarding negative values in violinplot
    # the "cut=0" truncates the data to the actual min and max values
    ax = sns.violinplot(x=df[xval],
                        y=df[yval], cut=0)
    ax.set_title(title)
    plt.xticks(rotation=15, ha='right')
    #plt.figtext(.0, -.1, str(kwargs)) # Append metainformation to the bottom of the chart (for export)
    ax.set_ylabel(ylabel)
    sns.despine(left=True)
    
    metainfo = get_metainfo_string(**kwargs)
    save_fig_in_subfolder(ax.figure, 'violin_' + yval + '_' + xval + metainfo + (start_yymm + '-' + end_yymm) if start_yymm is not None else('violin_' + yval + '_' + xval + metainfo + (start_yymmdd + '-' + end_yymmdd)) if start_yymmdd is not None else 'violin_' + yval + '_' + xval + metainfo + get_now_time_as_string())


def get_metainfo_string(**kwargs) -> str:
    result = ''
    for key, value in kwargs.items():
        value = value.replace('https://','')
        result = result + '-' + value
    return result


def _distance(a, b):
    if (a == b):
        return 0
    elif (a < 0) and (b < 0) or (a > 0) and (b > 0):
        if (a < b):
            return (abs(abs(a) - abs(b)))
        else:
            return -(abs(abs(a) - abs(b)))
    else:
        return math.copysign((abs(a) + abs(b)),b)
    
# From guid 11a
def define_heat_map(zval, yval, xval, df, agg, colours, title, start_yymm=None, end_yymm=None, start_yymmdd=None, end_yymmdd=None,vmin = 49.9, vmax = 50.1, unit='', figsize=(25,12), save=False):
    #limits
    hmin = 0
    hmax = 2359
    

    #if user date input: shorten data
    if start_yymm != None and end_yymm != None:
        df = df[df.YYMM >= int(start_yymm)]
        df = df[df.YYMM <= int(end_yymm)]
    if start_yymmdd != None and end_yymmdd != None:
        df = df[df.YYMMDD >= int(start_yymmdd)]
        df = df[df.YYMMDD <= int(end_yymmdd)]
    

    #create a pivot table to aggregate the data
    pivot_hm = pd.pivot_table (
            df,
            index      = [yval],
            columns    = [xval],
            values     = [zval], 
            #fill_value = 0,
            margins    = False,
            aggfunc    = [eval(agg)],
            dropna     = True  #keep null values
    )

    #Filter Data
    pivot_hm = pivot_hm[pivot_hm.index >= hmin]
    pivot_hm = pivot_hm[pivot_hm.index <= hmax]

    #Meta information
    """
    print('Tenant: ', url_widget.value)
    print('Source: ', wid_stream.value)
    print('Resolution: ', wid_reso.value)
    """
    if xval == 'YYMM':
        print('Data from: ', df['YYMM'].iloc[0], 'to: ', df['YYMM'].iloc[-1])
    elif xval == 'YYMMDD':
        print()
        #print('Data from: ', df['YYMMDD'].iloc[0], 'to: ', df['YYMMDD'].iloc[-1])
    else:
        print('Data from: ', df['DHOD'].iloc[0], 'to: ', df['DHOD'].iloc[-1])
    # create plot area
    fig, ax = plt.subplots(figsize=figsize)
    im = ax.imshow(
        pivot_hm, 
        vmin = vmin, 
        vmax = vmax, 
        cmap = colours, 
        origin ='lower', #desc y-axis values 
        aspect = 'auto',
        #interpolation='bilinear'  #comment out for no-interpolation
    ) 
  
    # Y AXIS rows
    pr_hm =  pivot_hm.shape[0]
    plt.yticks(np.arange(0,pr_hm), rotation=0)    
    
    ax.set_yticklabels(pivot_hm.index, fontsize = 18)

    # if to much yvalues shorten them
    if (len(pivot_hm) > 200):
        temp = []
        for indx, item in enumerate(pivot_hm.index):
            if indx % 60 == 0:
                temp.append(item)
            else:
                temp.append(None)
        ax.set_yticklabels(temp, fontsize=18)
    ax.set_ylabel(yval, fontsize=18)
    
    # add helper lines
    
    # differ sample rate
    constant = 60
    #for i in range(0,24):
    #    ax.axhline(i*constant, color='gray', linewidth=1 )
    diff = _distance(pivot_hm.shape[1], len(pivot_hm.columns.levels[2])) #DIRTY FIX for mod difference (set_xticks and set_xticklabels) remove this if breaks
    
    # X AXIS columns
    pc_hm = pivot_hm.shape[1] #number of 
    plt.xticks(np.arange(0, pc_hm+diff), rotation=90) 
    ax.set_xlabel(xval, fontsize=18)

    # Set costum tick labels TODO - change to 0.05 steps
    ax.set_xticklabels(pivot_hm.columns.levels[2], fontsize=18) 
    if len(pivot_hm.columns.levels[2]) > 100:    #create new ticks if more than 20 columns (x-axis)
        nw_tick_list = []
        for i, item in enumerate(pivot_hm.columns.levels[2]):
            if xval == 'YYMM' or xval == 'DHOD':
                if i % 2 == 0:
                    nw_tick_list.append(item)
                else:
                    nw_tick_list.append(None)
                    
            else:
                if i % 5 == 0:
                    nw_tick_list.append(item)
                else:
                    nw_tick_list.append(None)
        ax.set_xticklabels(nw_tick_list, fontsize = 18 ) #pivot_o.index)

    #define z-axis (colorbar)
    cbarlabel = zval + ' ' + unit
    cbar = ax.figure.colorbar(im, ax=ax) # , **cbar_kw)
    cbar.ax.set_ylabel(cbarlabel, rotation=90, va="bottom", fontsize=18, labelpad=30)
    #cbar.set_label(zval, labelpad=20) 
    
    #set labels
    #ax.grid(linewidth=1)
    fig.tight_layout()
    plt.xlabel(xval)
    plt.ylabel(yval)
    plt.title(title, fontsize=20)
    plt.tight_layout()
    plt.show()
    if save == True:
        save_fig_in_subfolder(fig, f'heatmap_{zval}_{get_now_time_as_string()}', 'svg') #svg to avoid losing details on heatmaps
    #return pivot_hm #only if you are interested in pivot
    
# From Guid 1
def double_y_axis_plot(df, xval, yvals, yval_labels, ylabel, y2label=None, y2vals=None, y2val_labels=None, *argv, **kwargs):
    
    #Meta information
    for key, value in kwargs.items():
        print(key + ' : ' + value)
    
    # args for costum description
    for arg in argv:
        print(arg)
    
    
    fig = plt.figure()

    # Convert timestamps to datetime for xaxis
    x_time_ms = np.array(df[xval])
    x_time_s = np.divide(x_time_ms,1000)
    len(x_time_s)
    x_time = [dt.datetime.fromtimestamp(ts) for ts in x_time_s]

    fig, ax1 = plt.subplots(figsize=(16,7))
    x_label = 'Date'
    ax1.set_xlabel(x_label)
    
    #set custom lable formatter for y-axis
    #from matplotlib.ticker import FuncFormatter
    #def kw(x, pos):
    #    return '%1.1f' % (x*1e-3)
    
    #formatter = FuncFormatter(kw)
    #ax1.yaxis.set_major_formatter(formatter)

    #plot left y-axis
    for i in range(len(yvals)):
        ax1.plot(x_time, df[yvals[i]], label=yval_labels[i])

    ax1.set_ylabel(ylabel)
    #ax1.set_ylim(0,1400)
    ax1.legend(frameon=True, loc='upper left', shadow=True) #put axis legends in left corner

    #plot right y-axis
    if y2vals != None:
        ax2 = ax1.twinx() #merge ax2 to first plot
        ax2.set_ylabel(y2label)
        # Start with color cycle on end of ax1 (to not have duplicate colors)
        ax2._get_lines.prop_cycler = ax1._get_lines.prop_cycler

        for i in range(len(y2vals)):
            ax2.plot(x_time, df[y2vals[i]], label=y2val_labels[i])

    #ax2.set_ylim(-20,70)
        ax2.legend(frameon=True, loc='upper right', shadow=True) #put axis legends in left corner

    #rotate x-ticks for better visability
    plt.setp(ax1.get_xticklabels(), rotation=15, horizontalalignment='right')
    
    fig.tight_layout()
    plt.legend()
    if y2label != None:
        plt.title((ylabel + ' vs ' + x_label + ' vs ' + y2label), fontsize=25)
    else:
        plt.title((ylabel + ' vs ' + x_label), fontsize=25)
    plt.show()
    
    save_fig_in_subfolder(fig, 'x_y_axis_plot' + get_now_time_as_string())
    
    
# From GUID 18-20
def pair_plot(df, variables, zval, *argv, **kwargs):
    
    #Meta information
    for key, value in kwargs.items():
        print(key + ' : ' + value)
    
    # args for costum description
    for arg in argv:
        print(arg)
    
    g = sns.pairplot(
        df,
        vars=variables,
        palette="hls",
        kind = "scatter",
        markers = 'o',
        hue = zval,
        plot_kws=dict(s=50, edgecolor="black", linewidth=0.1)
    )
    metainfo = get_metainfo_string(**kwargs)
    var_str = get_list_as_string(variables)
    
    save_fig_in_subfolder(g, 'pair_plot_' + '_' + metainfo + get_now_time_as_string())
    
def get_list_as_string(li): 
    res = ''
    for item in li:
        res += '_' + str(item)
    return res

def save_fig_in_subfolder(fig, fname, format=''):
    # TODO change path for linux
    save_dir = 'export/figures/'
    mkdir_p(save_dir)
    
    if format != '':
        fig.savefig(save_dir + fname + f'.{format}', format=format)
    else:
        fig.savefig(save_dir + fname + '.png')
    
    
def mkdir_p(mypath):
    '''Creates a directory. equivalent to using mkdir -p on the command line'''

    from errno import EEXIST
    from os import makedirs,path

    try:
        makedirs(mypath)
    except OSError as exc: # Python >2.5
        if exc.errno == EEXIST and path.isdir(mypath):
            pass
        else: raise
        

def get_now_time_as_string():
    
    utc_time = dt.datetime.utcnow() 
    return utc_time.strftime('_%Y_%m_%d-%H_%M_%S')


# -*- coding: utf-8 -*-
"""
Created on Wed Apr  8 16:47:07 2020
DRAFT
@author: jouanb
"""
from ctypes import*
import ctypes
import os


class GetProcessBufferServer():
    def __init__(self):
        try:
            # chemin=os.path.abspath(".\\Giutility\\giutility_x64.dll")
            chemin = "libGInsUtility.so"
            #gins_lib = "/home/gins/data/libGInsUtility.so"
            self.GINSDll = ctypes.cdll.LoadLibrary(chemin)
            print("64 bit DLL imported")
            print(chemin)
        except OSError:
            print("can not imported libGInsUtility.so")

        self.GINSDll._CD_eGateHighSpeedPort_GetPostProcessBufferInfo.argtypes = [
            c_int, c_char_p, c_size_t, c_char_p, c_size_t]
        self.GINSDll._CD_eGateHighSpeedPort_Init_PostProcessBuffer.argtypes = [
            c_char_p, POINTER(c_int), POINTER(c_int)]
        self.bufferID = ctypes.create_string_buffer(50)
        self.bufferName = ctypes.create_string_buffer(50)
        self.HCLIENT = c_int(-1)
        self.HCONNECTION = c_int(-1)

    def get_buffer_count(self):
        """create buffer"""
        self.count = self.GINSDll._CD_eGateHighSpeedPort_GetPostProcessBufferCount()
        return(self.count)

    def get_buffer_info(self, bufferIndex):
        ret = self.GINSDll._CD_eGateHighSpeedPort_GetPostProcessBufferInfo(
            bufferIndex, self.bufferID, len(self.bufferID), self.bufferName, len(self.bufferName))
        if ret != 0:
            print("error get_buffer_info")
        else:
            print("buffer index:", bufferIndex, "buffer name:", self.bufferName.value.decode(
                'UTF-8'), "buffer ID:", self.bufferID.value.decode('UTF-8'))
        return(self.bufferName.value.decode('UTF-8'), self.bufferID.value.decode('UTF-8'))

    def init_buffer_conn(self, sbufferID):
        self.sbufferID = sbufferID.encode('UTF-8')
        ret = self.GINSDll._CD_eGateHighSpeedPort_Init_PostProcessBuffer(
            self.sbufferID, byref(self.HCLIENT), byref(self.HCONNECTION))
        if(ret != 0):
            print("Init Connection Failed - ret:", ret)
